
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <utils.h>
#include <convexpoly.h>

/**
 * Polygon intersection code from:
 *
 * Haines, Eric, "Point in Polygon Strategies," Graphics Gems IV, ed.
 * Paul Heckbert, Academic Press, p. 24-46, 1994
 *
 * My reference is:
 * http://www.acm.org/pubs/tog/editors/erich/ptinpoly/
 */

typedef struct {
    double      vx, vy, c;     /* edge equation  vx*X + vy*Y + c = 0 */
    int         ext_flag;      /* TRUE == exterior edge of polygon */
} PlaneSet, *pPlaneSet ;

pPlaneSet PlaneSetup( double **pgon, int numverts );
int PlaneTest( pPlaneSet p_plane_set, int numverts, double point[2] );

ConvexPoly::ConvexPoly( const VertexList &vlist )
{
    Vector3 normal;
    int i;

    numpts = vlist.size();

    ptlist = new Vertex[ numpts ];
    for( i = 0; i < numpts; i++ ) {
        ptlist[ i ] = vlist[ i ];
    }

    Vector3 v1 = ptlist[ 1 ].pos.subtract( ptlist[ 0 ].pos ).normalize();
    Vector3 v2 = ptlist[ 2 ].pos.subtract( ptlist[ 0 ].pos ).normalize();
    normal = v1.cross( v2 ).normalize();

    double bx, by, bz;
    bx = fabs( normal.getX() );
    by = fabs( normal.getY() );
    bz = fabs( normal.getZ() );
    bigx = ( bx >= by ) && ( bx >= bz );
    bigy = ( by >= bx ) && ( by >= bz );
    bigz = ( bz >= bx ) && ( bz >= by );

    double **pgon = new double*[ numpts ];
    if( bigx ) {
        for( i = 0; i < numpts; i++ ) {
            pgon[ i ] = new double[ 2 ];
            pgon[ i ][ 0 ] = ptlist[ i ].pos.getY();
            pgon[ i ][ 1 ] = ptlist[ i ].pos.getZ();
        }
    } else if( bigy ) {
        for( i = 0; i < numpts; i++ ) {
            pgon[ i ] = new double[ 2 ];
            pgon[ i ][ 0 ] = ptlist[ i ].pos.getX();
            pgon[ i ][ 1 ] = ptlist[ i ].pos.getZ();
        }
    } else {
        for( i = 0; i < numpts; i++ ) {
            pgon[ i ] = new double[ 2 ];
            pgon[ i ][ 0 ] = ptlist[ i ].pos.getX();
            pgon[ i ][ 1 ] = ptlist[ i ].pos.getY();
        }
    }
    pset = PlaneSetup( pgon, numpts );

    min0 = HUGE_VAL; max0 = -HUGE_VAL;
    min1 = HUGE_VAL; max1 = -HUGE_VAL;
    for( i = 0; i < numpts; i++ ) {
        if( pgon[ i ][ 0 ] < min0 ) min0 = pgon[ i ][ 0 ];
        if( pgon[ i ][ 0 ] > max0 ) max0 = pgon[ i ][ 0 ];
        if( pgon[ i ][ 1 ] < min1 ) min1 = pgon[ i ][ 1 ];
        if( pgon[ i ][ 1 ] > max1 ) max1 = pgon[ i ][ 1 ];
    }

    a = normal.getX();
    b = normal.getY();
    c = normal.getZ();
    d =   ( -a * ptlist[ 0 ].pos.getX() )
        - (  b * ptlist[ 0 ].pos.getY() )
        - (  c * ptlist[ 0 ].pos.getZ() );
}

ConvexPoly::~ConvexPoly( void )
{
    delete[] ptlist;
}

BoundingBox ConvexPoly::getBoundingBox( void ) const
{
    BoundingBox b;
    int i;

    b.minx = HUGE_VAL; b.maxx = -HUGE_VAL;
    b.miny = HUGE_VAL; b.maxy = -HUGE_VAL;
    b.minz = HUGE_VAL; b.maxz = -HUGE_VAL;

    for( i = 0; i < numpts; i++ ) {
        if( ptlist[ i ].pos.getX() < b.minx ) b.minx = ptlist[ i ].pos.getX();
        if( ptlist[ i ].pos.getX() > b.maxx ) b.maxx = ptlist[ i ].pos.getX();
        if( ptlist[ i ].pos.getY() < b.miny ) b.miny = ptlist[ i ].pos.getY();
        if( ptlist[ i ].pos.getY() > b.maxy ) b.maxy = ptlist[ i ].pos.getY();
        if( ptlist[ i ].pos.getZ() < b.minz ) b.minz = ptlist[ i ].pos.getZ();
        if( ptlist[ i ].pos.getZ() > b.maxz ) b.maxz = ptlist[ i ].pos.getZ();
    }

    return b;
}

PrimitiveList ConvexPoly::foldModel( const Matrix4x4 &xform ) const
{
    PrimitiveList pl;
    VertexList newlist;
    int i;

    for( i = 0; i < numpts; i++ ) {
        Vertex newvert = ptlist[ i ];
        newvert.pos = multiplyPointByMatrix( xform, ptlist[ i ].pos );
        if( newvert.hasnormal ) {
            newvert.normal = multiplyVectorByMatrix( xform, ptlist[ i ].normal );
            newvert.normal = newvert.normal.normalize();
        }
        newlist.push_back( newvert );
    }

    ConvexPoly *c = new ConvexPoly( newlist );
    c->setMaterial( getMaterial() );
    c->setSampleRate( getSampleRate() );
    pl.push_back( c );
    return pl;
}

VertexList ConvexPoly::foldToVerticies( const Matrix4x4 &xform ) const
{
    VertexList newlist;
    int i;

    for( i = 0; i < numpts; i++ ) {
        Vertex newvert = ptlist[ i ];
        newvert.pos = multiplyPointByMatrix( xform, ptlist[ i ].pos );
        if( newvert.hasnormal ) {
            newvert.normal = multiplyVectorByMatrix( xform, ptlist[ i ].normal );
            newvert.normal = newvert.normal.normalize();
        }
        newlist.push_back( newvert );
    }
    return newlist;
}

bool ConvexPoly::planeIntersect( const Point3 &origin, const Vector3 &dir,
                                 double *t, Vector3 *norm ) const
{
    double bot = (  ( a * dir.getX() )
                  + ( b * dir.getY() )
                  + ( c * dir.getZ() ) );
    if( threshEqual( 0.0, bot ) ) return false;

    double top = ( ( -a * origin.getX() )
                 - (  b * origin.getY() )
                 - (  c * origin.getZ() ) - d );
    if( ( top < 0.0 && bot > 0.0 ) || ( top > 0.0 && bot < 0.0 ) ) return false;
    *t = top / bot;
    if( *t <= 0.0 ) return false;

    *norm = Vector3( a, b, c );
    if( dir.dotProduct( *norm ) >= 0.0 ) {
        *norm = (*norm).scalarMult( -1.0 );
    }

    return true;
}

bool ConvexPoly::closestIntersect( const Point3 &origin, const Vector3 &dir,
                                   double *t, Vector3 *norm, Material **mat ) const
{
    Vector3 v1, v2;
    Point3 intpoint;
    double intpointd[2];
    double curt;

    if( !planeIntersect( origin, dir, &curt, norm ) ) return false;
    if( curt > *t ) return false;
    *t = curt;

    intpoint = origin.add( dir.scalarMult( *t ) );

    if( bigx ) {
        intpointd[ 0 ] = intpoint.getY();
        intpointd[ 1 ] = intpoint.getZ();
    } else if( bigy ) {
        intpointd[ 0 ] = intpoint.getX();
        intpointd[ 1 ] = intpoint.getZ();
    } else {
        intpointd[ 0 ] = intpoint.getX();
        intpointd[ 1 ] = intpoint.getY();
    }

    if( intpointd[ 0 ] < min0 || intpointd[ 0 ] > max0 ) return false;
    if( intpointd[ 1 ] < min1 || intpointd[ 1 ] > max1 ) return false;

    if( PlaneTest( (pPlaneSet) pset, numpts, intpointd ) ) {
        *mat = getMaterial();
        return true;
    }
    return false;
}

/* http://vcg.iei.pi.cnr.it/pointintetra.htm */
Point3 ConvexPoly::getRandomLightPosition( const Point3 &closest,
                                           double *probability ) const
{
    if( numpts != 4 ) return Point3::origin;

    double s = ( (double) rand() / (double) RAND_MAX );
    double t = ( (double) rand() / (double) RAND_MAX );
    double u = ( (double) rand() / (double) RAND_MAX );

    if( s + t > 1.0 ) {
        /* Cut'n fold the cube into a prism. */
        s = 1.0 - s;
        t = 1.0 - t;
    }

    if( t + u > 1.0 ) {
        /* Cut'n fold the prism into a tetrahedron. */
        double tmp = u;
        u = 1.0 - s - t;
        t = 1.0 - tmp;
    } else if( s + t + u > 1.0 ) {
        double tmp = u;
        u = s + t + u - 1.0;
        s = 1.0 - t - tmp;
    }

    /* a,s,t,u are the barycentric coordinates of the random point. */
    double v = 1.0 - s - t - u;
    Point3 ret;
    ret = Point3::origin.add( ptlist[ 0 ].pos.subtractOrigin().scalarMult( v ) );
    ret = ret.add( ptlist[ 1 ].pos.subtractOrigin().scalarMult( s ) );
    ret = ret.add( ptlist[ 2 ].pos.subtractOrigin().scalarMult( t ) );
    ret = ret.add( ptlist[ 3 ].pos.subtractOrigin().scalarMult( u ) );
    ret = ret.add( closest.subtract( ret ).scalarMult( epsilon ) );

    Vector3 norm = Vector3( a, b, c ).normalize();
    Vector3 vect = closest.subtract( ret ).normalize();
    *probability = norm.dotProduct( vect );
    return ret;
}

/* ======= Triangle half-plane algorithm ================================== */

/* Split the polygon into a fan of triangles and for each triangle test if
 * the point is inside of the three half-planes formed by the triangle's edges.
 *
 * Call setup with 2D polygon _pgon_ with _numverts_ number of vertices,
 * which returns a pointer to a plane set array.
 * Call testing procedure with a pointer to this array, _numverts_, and
 * test point _point_, returns 1 if inside, 0 if outside.
 * Call cleanup with pointer to plane set array to free space.
 *
 * SORT and CONVEX can be defined for this test.
 */


/* Size sorting structure for half-planes */
typedef struct {
    double      size ;
    pPlaneSet   pps ;
} SizePlanePair, *pSizePlanePair ;

int CompareSizePlanePairs( const void *a1, const void *a2 )
{
    pSizePlanePair p_sp0 = (pSizePlanePair) a1;
    pSizePlanePair p_sp1 = (pSizePlanePair) a2;
    if ( p_sp0->size == p_sp1->size ) {
        return( 0 ) ;
    } else {
        return( p_sp0->size > p_sp1->size ? -1 : 1 ) ;
    }
}

/* split polygons along set of x axes - call preprocess once */
pPlaneSet PlaneSetup( double **pgon, int numverts )
{
    int i, p1, p2;
    double tx, ty, vx0, vy0;
    pPlaneSet pps, pps_return;
    double len[3], len_temp;
    int j;
    PlaneSet ps_temp;
    pPlaneSet pps_new;
    pSizePlanePair p_size_pair;

    pps = pps_return = (pPlaneSet)malloc( 3 * (numverts-2) * sizeof( PlaneSet ));
    // MALLOC_CHECK( pps ) ;
    p_size_pair = (pSizePlanePair)malloc( (numverts-2) * sizeof( SizePlanePair ) );
    // MALLOC_CHECK( p_size_pair ) ;

    vx0 = pgon[0][0] ;
    vy0 = pgon[0][1] ;

    for ( p1 = 1, p2 = 2 ; p2 < numverts ; p1++, p2++ ) {
        pps->vx = vy0 - pgon[p1][1] ;
        pps->vy = pgon[p1][0] - vx0 ;
        pps->c = pps->vx * vx0 + pps->vy * vy0 ;
        len[0] = pps->vx * pps->vx + pps->vy * pps->vy ;
        pps->ext_flag = ( p1 == 1 );

        /* Sort triangles by areas, so compute (twice) the area here */
        p_size_pair[p1-1].pps = pps ;
        p_size_pair[p1-1].size =
                        ( pgon[0][0] * pgon[p1][1] ) +
                        ( pgon[p1][0] * pgon[p2][1] ) +
                        ( pgon[p2][0] * pgon[0][1] ) -
                        ( pgon[p1][0] * pgon[0][1] ) -
                        ( pgon[p2][0] * pgon[p1][1] ) -
                        ( pgon[0][0] * pgon[p2][1] ) ;
        pps++ ;
        pps->vx = pgon[p1][1] - pgon[p2][1] ;
        pps->vy = pgon[p2][0] - pgon[p1][0] ;
        pps->c = pps->vx * pgon[p1][0] + pps->vy * pgon[p1][1] ;
        len[1] = pps->vx * pps->vx + pps->vy * pps->vy ;
        pps->ext_flag = 1;
        pps++ ;
        pps->vx = pgon[p2][1] - vy0 ;
        pps->vy = vx0 - pgon[p2][0] ;
        pps->c = pps->vx * pgon[p2][0] + pps->vy * pgon[p2][1] ;
        len[2] = pps->vx * pps->vx + pps->vy * pps->vy ;
        pps->ext_flag = ( p2 == numverts-1 );

        /* find an average point which must be inside of the triangle */
        tx = ( vx0 + pgon[p1][0] + pgon[p2][0] ) / 3.0 ;
        ty = ( vy0 + pgon[p1][1] + pgon[p2][1] ) / 3.0 ;

        /* check sense and reverse if test point is not thought to be inside
         * first triangle
         */
        if ( pps->vx * tx + pps->vy * ty >= pps->c ) {
            /* back up to start of plane set */
            pps -= 2 ;
            /* point is thought to be outside, so reverse sense of edge
             * normals so that it is correctly considered inside.
             */
            for ( i = 0 ; i < 3 ; i++ ) {
                pps->vx = -pps->vx ;
                pps->vy = -pps->vy ;
                pps->c  = -pps->c ;
                pps++ ;
            }
        } else {
            pps++ ;
        }

        /* sort the planes based on the edge lengths */
        pps -= 3 ;
        for ( i = 0 ; i < 2 ; i++ ) {
            for ( j = i+1 ; j < 3 ; j++ ) {
                if ( len[i] < len[j] ) {
                    ps_temp = pps[i] ;
                    pps[i] = pps[j] ;
                    pps[j] = ps_temp ;
                    len_temp = len[i] ;
                    len[i] = len[j] ;
                    len[j] = len_temp ;
                }
            }
        }
        pps += 3 ;
    }

    /* sort the triangles based on their areas */
    qsort( p_size_pair, numverts-2,
            sizeof( SizePlanePair ), CompareSizePlanePairs ) ;

    /* make the plane sets match the sorted order */
    for ( i = 0, pps = pps_return
        ; i < numverts-2
        ; i++ ) {

        pps_new = p_size_pair[i].pps ;
        for ( j = 0 ; j < 3 ; j++, pps++, pps_new++ ) {
            ps_temp = *pps ;
            *pps = *pps_new ;
            *pps_new = ps_temp ;
        }
    }
    free( p_size_pair ) ;

    return( pps_return ) ;
}


/* check point for inside of three "planes" formed by triangle edges */
int PlaneTest( pPlaneSet p_plane_set, int numverts, double point[2] )
{
    register pPlaneSet ps;
    register int p2;
    register double tx, ty;

    tx = point[0];
    ty = point[1];

    for( ps = p_plane_set, p2 = numverts-1; --p2; ) {

        if( ps->vx * tx + ps->vy * ty < ps->c ) {
            ps++;
            if( ps->vx * tx + ps->vy * ty < ps->c ) {
                ps++;
                /* note: we make the third edge have a slightly different
                 * equality condition, since this third edge is in fact
                 * the next triangle's first edge.  Not fool-proof, but
                 * it doesn't hurt (better would be to keep track of the
                 * triangle's area sign so we would know which kind of
                 * triangle this is).  Note that edge sorting nullifies
                 * this special inequality, too.
                 */
                if( ps->vx * tx + ps->vy * ty <= ps->c ) {
                    /* point is inside polygon */
                    return 1;
                }

                /* check if outside exterior edge */
                else if( ps->ext_flag ) return 0;

                ps++;
            } else {
                /* check if outside exterior edge */
                if( ps->ext_flag ) return 0;

                /* get past last two plane tests */
                ps += 2;
            }
        } else {
            /* check if outside exterior edge */
            if( ps->ext_flag ) return 0;

            /* get past all three plane tests */
            ps += 3;
        }
    }

    /* for convex, if we make it to here, all triangles were missed */
    return 0;
}

void PlaneCleanup( pPlaneSet p_plane_set )
{
    free( p_plane_set ) ;
}

