
#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include "vector3.h"
#include "model.h"
#include "sceneparser.h"
#include "pnginput.h"
#include "pngtexture.h"
#include "cooktorrence.h"

/**
 * Cook-Torrence Parameters.
 */
static double diffuse_r, diffuse_g, diffuse_b,
              specular_r, specular_g, specular_b,
              ct_m, ct_ks, ct_kd;
static double s_diffuse_r, s_diffuse_g, s_diffuse_b,
              s_specular_r, s_specular_g, s_specular_b,
              s_ct_m, s_ct_ks, s_ct_kd;
static int ct_size = 256;


static Model *m;
static Vector3 camera_position( 0.0, 0.0, -2.0 );
static Vector3 camera_up( 0.0, 1.0, 0.0 );
static Vector3 camera_direction = Vector3( 0.0, 0.0, 3.6 ).normalize();
static Vector3 light_position( 0.0, 0.0, -3.6 );
static Vector3 light_up( 0.0, 1.0, 0.0 );

static CookTorrence *ct;
static CookTorrence *ct2;

static TextureData *mask;
static TextureData *invmask;

static void init( void )
{
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
    glEnable( GL_LIGHTING );
    glEnable( GL_COLOR_MATERIAL );
    glEnable( GL_ALPHA_TEST );
    glEnable( GL_DEPTH_TEST );
    glEnable( GL_BLEND );
    glEnable( GL_CULL_FACE );

    glDepthFunc( GL_LEQUAL );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    assert( glGetError() == GL_NO_ERROR );

    GLfloat lightDiffuse[]= { 1.0f, 1.0f, 1.0f, 1.0f };
    glLightfv( GL_LIGHT0, GL_DIFFUSE, (const GLfloat *) lightDiffuse );
}

static void display( void )
{
    /* Clear the colour and depth. */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    /* Setup camera position. */
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    gluLookAt( camera_position.getX(),
               camera_position.getY(),
               camera_position.getZ(),
               camera_position.getX() + camera_direction.getX(),
               camera_position.getY() + camera_direction.getY(),
               camera_position.getZ() + camera_direction.getZ(),
               camera_up.getX(), camera_up.getY(), camera_up.getZ() );

    /* Setup our light position. */
    GLfloat lightPosition[4];
    lightPosition[ 0 ] = light_position.getX();
    lightPosition[ 1 ] = light_position.getY();
    lightPosition[ 2 ] = light_position.getZ();
    lightPosition[ 3 ] = 1.0;
    glLightfv( GL_LIGHT0, GL_POSITION, (const GLfloat *) lightPosition );
    glEnable( GL_LIGHT0 );


    glEnable( GL_TEXTURE_2D );



    /**
     * Show off our pretty Cook-Torrence model, part 1!
     */

    /**
     * First do the first pass of the first CT model.
     */
    glActiveTextureARB( GL_TEXTURE0_ARB );
    ct->getTexture1()->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glActiveTextureARB( GL_TEXTURE1_ARB );
    mask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glColor4f( 1.0, 1.0, 1.0, 1.0 );
    m->drawModelWithLighting_vh_nh( Point3::origin.add( camera_position ),
                                    Point3::origin.add( light_position ) );


    /**
     * Next, add in the first pass of the second CT model.
     */
    glActiveTextureARB( GL_TEXTURE0_ARB );
    ct2->getTexture1()->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glActiveTextureARB( GL_TEXTURE1_ARB );
    invmask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glColor4f( 1.0, 1.0, 1.0, 1.0 );
    m->drawModelWithLighting_vh_nh( Point3::origin.add( camera_position ),
                                    Point3::origin.add( light_position ) );


    /**
     * Second pass of the first model.
     */

    glActiveTextureARB( GL_TEXTURE0_ARB );
    ct->getTexture2()->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glActiveTextureARB( GL_TEXTURE1_ARB );
    mask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glBlendFunc( GL_DST_COLOR, GL_ONE );
    m->drawModelWithLighting_nl_nv( Point3::origin.add( camera_position ),
                                    Point3::origin.add( light_position ) );


    /**
     * Second pass of the second model.
     */
    glActiveTextureARB( GL_TEXTURE0_ARB );
    ct2->getTexture2()->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glActiveTextureARB( GL_TEXTURE1_ARB );
    invmask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glBlendFunc( GL_DST_COLOR, GL_ONE );
    m->drawModelWithLighting_nl_nv( Point3::origin.add( camera_position ),
                                    Point3::origin.add( light_position ) );


    /**
     * Add in diffuse terms.
     */
    glActiveTextureARB( GL_TEXTURE0_ARB );
    mask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glActiveTextureARB( GL_TEXTURE1_ARB );
    mask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glColor4f( diffuse_r, diffuse_g, diffuse_b, ct_kd );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE );
    m->drawModel();


    glActiveTextureARB( GL_TEXTURE0_ARB );
    invmask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glActiveTextureARB( GL_TEXTURE1_ARB );
    invmask->bindTexture();
    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    glColor4f( s_diffuse_r, s_diffuse_g, s_diffuse_b, s_ct_kd );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE );
    m->drawModel();

    glFlush();
    glutSwapBuffers();
}

static void reshape( int w, int h )
{
    glViewport( 0, 0, (GLsizei) w, (GLsizei) h );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 30.0, (GLfloat) w / (GLfloat) h, 0.01, 30.0 );
    glMatrixMode( GL_MODELVIEW );
}

static void keyboard( unsigned char key, int x, int y )
{
    switch( key ) {
        case 27: exit( 0 ); break;
        case 'q': light_position = light_position.add( Vector3( 0.5, 0.0, 0.0 ) );
                  glutPostRedisplay();
                  break;
        case 'a': light_position = light_position.add( Vector3( -0.5, 0.0, 0.0 ) );
                  glutPostRedisplay();
                  break;
        case 'w': light_position = light_position.add( Vector3( 0.0, 0.5, 0.0 ) );
                  glutPostRedisplay();
                  break;
        case 's': light_position = light_position.add( Vector3( 0.0, -0.5, 0.0 ) );
                  glutPostRedisplay();
                  break;
        case 'e': light_position = light_position.add( Vector3( 0.0, 0.0, 0.5 ) );
                  glutPostRedisplay();
                  break;
        case 'd': light_position = light_position.add( Vector3( 0.0, 0.0, -0.5 ) );
                  glutPostRedisplay();
                  break;

        case 'r':
            camera_position = camera_position.rotate( 0.1, Vector3( 0.0, 1.0, 0.0 ) );
            camera_direction = camera_position.scalarMult( -1.0 ).normalize();
            glutPostRedisplay();
            break;
        case 'f':
            camera_position = camera_position.rotate( -0.1, Vector3( 0.0, 1.0, 0.0 ) );
            camera_direction = camera_position.scalarMult( -1.0 ).normalize();
            glutPostRedisplay();
            break;
        case 'o':
            ct_m += 0.1;
            if( ct_m > 1.0 ) ct_m = 1.0;
            fprintf( stderr, "q3demo: m value now %f\n", ct_m );
            ct->rebuild( diffuse_r, diffuse_g, diffuse_b,
                         specular_r, specular_g, specular_b,
                         ct_kd, ct_ks, ct_m, ct_size );
            fprintf( stderr, "q3demo: done rebuilding.\n" );
            glutPostRedisplay();
            break;
        case 'l':
            ct_m -= 0.1;
            if( ct_m < 0.1 ) ct_m = 0.1;
            fprintf( stderr, "q3demo: m value now %f\n", ct_m );
            ct->rebuild( diffuse_r, diffuse_g, diffuse_b,
                         specular_r, specular_g, specular_b,
                         ct_kd, ct_ks, ct_m, ct_size );
            fprintf( stderr, "q3demo: done rebuilding.\n" );
            glutPostRedisplay();
            break;
        case 'i':
            s_ct_m += 0.1;
            if( s_ct_m > 1.0 ) s_ct_m = 1.0;
            fprintf( stderr, "q3demo: s_m value now %f\n", s_ct_m );
            ct2->rebuild( s_diffuse_r, s_diffuse_g, s_diffuse_b,
                          s_specular_r, s_specular_g, s_specular_b,
                          s_ct_kd, s_ct_ks, s_ct_m, ct_size );
            fprintf( stderr, "q3demo: done rebuilding.\n" );
            glutPostRedisplay();
            break;
        case 'k':
            s_ct_m -= 0.1;
            if( s_ct_m < 0.1 ) s_ct_m = 0.1;
            fprintf( stderr, "q3demo: s_m value now %f\n", s_ct_m );
            ct2->rebuild( s_diffuse_r, s_diffuse_g, s_diffuse_b,
                          s_specular_r, s_specular_g, s_specular_b,
                          s_ct_kd, s_ct_ks, s_ct_m, ct_size );
            fprintf( stderr, "q3demo: done rebuilding.\n" );
            glutPostRedisplay();
            break;

        default: fprintf( stderr, "normal: %d\n", key ); break;
    }
}

static void special_keyboard( int key, int x, int y )
{
    switch( key ) {
        case 101: /* Up. */
            camera_position = camera_position.add( camera_direction.scalarMult( 0.1 ) );
            glutPostRedisplay();
            break;
        case 103: /* Down. */
            camera_position = camera_position.add( camera_direction.scalarMult( -0.1 ) );
            glutPostRedisplay();
            break;
        case 100: /* Left. */
            camera_direction = camera_direction.rotate( 0.1, Vector3( 0.0, 1.0, 0.0 ) ).normalize();
            glutPostRedisplay();
            break;
        case 102: /* Right. */
            camera_direction = camera_direction.rotate( -0.1, Vector3( 0.0, 1.0, 0.0 ) ).normalize();
            glutPostRedisplay();
            break;
        default: fprintf( stderr, "special: %d\n", key ); break;
    }
}

static int last_x, last_y;
static bool move_camera = true;
static bool zoom_camera = false;

static void mouse_motion( int x, int y )
{
    if( move_camera ) {
        if( zoom_camera ) {
            camera_position = camera_position.rotate( (last_x - x) * 0.01,
                                                      camera_up );
            camera_direction = camera_position.scalarMult( -1.0 ).normalize();
            camera_position = camera_position.add( camera_direction.scalarMult( (last_y - y) * 0.01 ) );
        } else {
            Vector3 perp = camera_position.normalize().cross( camera_up.normalize() );

            camera_up = camera_up.rotate( (y - last_y) * 0.01, perp );
            camera_position = camera_position.rotate( (y - last_y) * 0.01, perp );
            camera_position = camera_position.rotate( (last_x - x) * 0.01,
                                                      camera_up );
            camera_direction = camera_position.scalarMult( -1.0 ).normalize();
        }
    } else {
        Vector3 perp = light_position.normalize().cross( light_up.normalize() );

        light_position = light_position.rotate( (x - last_x) * 0.01,
                                                light_up );
        light_up = light_up.rotate( (last_y - y) * 0.01, perp );
        light_position = light_position.rotate( (last_y - y) * 0.01, perp );
    }

    last_x = x;
    last_y = y;
    glutPostRedisplay();
}

static void mouse_button( int button, int state, int x, int y )
{
    last_x = x;
    last_y = y;
}

static void menu_func( int value )
{
    move_camera = ( value == 1 || value == 2 );
    zoom_camera = ( value == 2 );
}

int main( int argc, char **argv )
{
    if( argc < 4 ) {
        fprintf( stderr, "Usage: %s <model file.txt> "
                         "<mask.png> <inverse of mask.png>\n", argv[ 0 ] );
        return 1;
    }

    SceneParser p( argv[ 1 ] );
    Node *teapot = p.getNode( "scene" );
    m = new Model( teapot->foldToVerticies( identityMatrix() ) );
    m->centreModel();

    glutInit( &argc, argv );
    glutInitDisplayMode( GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE );
    glutInitWindowSize( 512, 512 );
    glutInitWindowPosition( 100, 100 );
    glutCreateWindow( argv[ 0 ] );
    init();

    glutMotionFunc( mouse_motion );
    glutMouseFunc( mouse_button );

    glutCreateMenu( menu_func );
    glutAddMenuEntry( "Move Camera", 1 );
    glutAddMenuEntry( "Zoom Camera", 2 );
    glutAddMenuEntry( "Move Light", 3 );
    glutAttachMenu( GLUT_RIGHT_BUTTON );

    glutDisplayFunc( display );
    glutReshapeFunc( reshape );
    glutKeyboardFunc( keyboard );
    glutSpecialFunc( special_keyboard );

    diffuse_r = 0.0112648;
    diffuse_g = 0.725164;
    diffuse_b = 0.990648;
    specular_r = 0.106065;
    specular_g = 0.525164;
    specular_b = 0.955802;
    ct_kd = 0.5;
    ct_ks = 0.1;
    ct_m = 0.8;

    s_diffuse_r = 0.75164;
    s_diffuse_g = 0.60648;
    s_diffuse_b = 0.22648;
    s_specular_r = 0.728281;
    s_specular_g = 0.655802;
    s_specular_b = 0.166065;
    s_ct_kd = 0.5;
    s_ct_ks = 0.5;
    s_ct_m = 0.2;

    ct = new CookTorrence( diffuse_r, diffuse_g, diffuse_b,
                           specular_r, specular_g, specular_b,
                           ct_kd, ct_ks, ct_m, ct_size );
    ct2 = new CookTorrence( s_diffuse_r, s_diffuse_g, s_diffuse_b,
                            s_specular_r, s_specular_g, s_specular_b,
                            s_ct_kd, s_ct_ks, s_ct_m, ct_size );
    mask = new PNGTexture( argv[ 2 ] );
    invmask = new PNGTexture( argv[ 3 ] );
    if( !mask->wasSuccessful() || !invmask->wasSuccessful() ) {
        fprintf( stderr, "q3demo: Can't load masks!\n" );
        return 0;
    }

    glutMainLoop();
    return 0;
}

