
#include <stl.h>
#include <string>
#include <string.h>
#include <stdio.h>
#include <node.h>
#include <vertexlist.h>
#include <primitive.h>
#include <convexpoly.h>
#include <triangle.h>
#include <pointlight.h>
#include <sceneparser.h>

using namespace std;

class SceneReader
{
public:
    SceneReader( const char *filename )
        : done( false )
    {
        infile.open( filename );
        if( infile.bad() ) {
            fprintf( stderr, "sceneparser: Can't open %s.\n", filename );
            done = true;
        }
    }

    string getNextToken( void )
    {
        string curt;
        if( infile.eof() ) { done = true; }
        infile >> curt;
        if( curt == "/*" ) {
            while( !infile.eof() ) {
                infile >> curt;
                if( infile.eof() ) break;
                if( curt == "*/" ) break;
            }
        }
        return curt;
    }

    bool isDone( void ) const { return done; }

private:
    ifstream infile;
    bool done;
};

SceneParser::SceneParser( const char *filename )
    : width( 640 ), height( 480 ), gridwidth( 10 ),
      supersample( false ), superdepth( 2 ), lmode( None ),
      fatal( false )
{
    string curtoken;

    reader = new SceneReader( filename );
    if( !reader ) { fatal = true; return; }
    if( reader->isDone() ) fatal = true;

    while( !reader->isDone() ) {
        curtoken = reader->getNextToken();

        if( curtoken == "material" ) {
            parseMaterial();
        } else if( curtoken == "ashikhmin" ) {
            lmode = Ashikhmin;
        } else if( curtoken == "phong" ) {
            lmode = Phong;
        } else if( curtoken == "node" ) {
            nodes.push_back( parseNode() );
        } else if( curtoken == "camera" ) {
            parseCamera();
        } else if( curtoken == "width" ) {
            width = atoi( reader->getNextToken().c_str() );
        } else if( curtoken == "gridwidth" ) {
            gridwidth = atoi( reader->getNextToken().c_str() );
        } else if( curtoken == "height" ) {
            height = atoi( reader->getNextToken().c_str() );
        } else if( curtoken == "supersample" ) {
            supersample = true;
            superdepth = atoi( reader->getNextToken().c_str() );
        } else if( curtoken == "include" ) {
            /* Create a new sceneparser. */
            SceneParser incfile( reader->getNextToken().c_str() );

            /* Grab the nodes and materials. */
            nodes.merge( incfile.nodes );
            materials.merge( incfile.materials );
        }
    }
}

SceneParser::~SceneParser( void )
{
}

static inline void readTriple( SceneReader *r, double &x, double &y, double &z )
{
    x = atof( r->getNextToken().c_str() );
    y = atof( r->getNextToken().c_str() );
    z = atof( r->getNextToken().c_str() );
}

void SceneParser::parseMaterial( void )
{
    Material *m = new Material;
    string curtoken;

    /* Read material name. */
    m->setName( reader->getNextToken().c_str() );

    while( !reader->isDone() ) {
        curtoken = reader->getNextToken();

        if( curtoken == "endmaterial" ) break;
        if( curtoken == "rgb" ) {
            double r, g, b;
            readTriple( reader, r, g, b );
            m->setDiffuseColour( PixelXYZ::createPixelRGB( r, g, b ) );
        } else if( curtoken == "xyz" ) {
            double x, y, z;
            readTriple( reader, x, y, z );
            m->setDiffuseColour( PixelXYZ( x, y, z ) );
        } else if( curtoken == "spec-rgb" ) {
            double r, g, b;
            readTriple( reader, r, g, b );
            m->setSpecularColour( PixelXYZ::createPixelRGB( r, g, b ) );
        } else if( curtoken == "spec-xyz" ) {
            double x, y, z;
            readTriple( reader, x, y, z );
            m->setSpecularColour( PixelXYZ( x, y, z ) );
        } else if( curtoken == "ks" ) {
            m->setKs( atof( reader->getNextToken().c_str() ) );
        } else if( curtoken == "kd" ) {
            m->setKd( atof( reader->getNextToken().c_str() ) );
        } else if( curtoken == "nu" ) {
            m->setNu( atof( reader->getNextToken().c_str() ) );
        } else if( curtoken == "nv" ) {
            m->setNv( atof( reader->getNextToken().c_str() ) );
        } else if( curtoken == "luminous" ) {
            m->setLuminous( true );
        }
    }
    materials.push_back( m );
}

void SceneParser::parseCamera( void )
{
    Point3 pos( 0.0, 0.0, 40.0 );
    Vector3 up( 0.0, 1.0, 0.0 );
    Vector3 fwd( 0.0, 0.0, -1.0 );
    double fov = 30.0;
    string curtoken;

    while( !reader->isDone() ) {
        curtoken = reader->getNextToken();
        if( curtoken == "endcamera" ) break;
        if( curtoken == "up" ) {
            double x, y, z;
            readTriple( reader, x, y, z );
            up = Vector3( x, y, z );
        } else if( curtoken == "forward" ) {
            double x, y, z;
            readTriple( reader, x, y, z );
            fwd = Vector3( x, y, z );
        } else if( curtoken == "eyepos" ) {
            double x, y, z;
            readTriple( reader, x, y, z );
            pos = Point3( x, y, z );
        } else if( curtoken == "fov" ) {
            fov = atof( reader->getNextToken().c_str() );
        }
    }

    camera = Camera( pos, up, fwd, fov );
}

Material *SceneParser::getMaterial( const char *name ) const
{
    list<Material *>::const_iterator it;

    for( it = materials.begin(); it != materials.end(); it++ ) {
        if( strcmp( name, (*it)->getName() ) == 0 ) {
            return (*it);
        }
    }

    return 0;
}

Node *SceneParser::getNode( const char *name ) const
{
    list<Node *>::const_iterator it;

    for( it = nodes.begin(); it != nodes.end(); it++ ) {
        if( strcmp( name, (*it)->getName() ) == 0 ) {
            return (*it);
        }
    }

    return 0;
}

struct Polygon
{
    vector<int> verticies;
};

void SceneParser::readPolygons( Node *n, VertexList &vlist )
{
    VertexList::iterator vit;
    list<Polygon>::iterator pit;
    list<Polygon> polylist;
    string curtoken;

    (void) reader->getNextToken(); // read {
    while( !reader->isDone() ) {
        Polygon curgon;

        curtoken = reader->getNextToken();
        if( curtoken == "}" ) break;

        // Read until end of polygon.
        while( !reader->isDone() ) {
            curtoken = reader->getNextToken();
            if( curtoken == "}" ) break;
            curgon.verticies.push_back( atoi( curtoken.c_str() ) - 1 );
        }
        polylist.push_back( curgon );
    }

    /* We have all the polygons, add up their normals. */
    for( pit = polylist.begin(); pit != polylist.end(); pit++ ) {
        int pt0 = (*pit).verticies[ 0 ];
        int pt1 = (*pit).verticies[ 1 ];
        int pt2 = (*pit).verticies[ 2 ];

        Vector3 edge1 = vlist[ pt1 ].pos.subtract( vlist[ pt0 ].pos );
        Vector3 edge2 = vlist[ pt2 ].pos.subtract( vlist[ pt0 ].pos );

        Vector3 v1 = edge1.normalize();
        Vector3 v2 = edge2.normalize();
        Vector3 curnorm = v1.cross( v2 ).normalize();

        vlist[ pt0 ].normal = vlist[ pt0 ].normal.add( curnorm );
        vlist[ pt1 ].normal = vlist[ pt1 ].normal.add( curnorm );
        vlist[ pt2 ].normal = vlist[ pt2 ].normal.add( curnorm );
    }

    /* Cool!  Now, normalize all the normals. */
    for( vit = vlist.begin(); vit != vlist.end(); vit++ ) {
        (*vit).normal = (*vit).normal.normalize();
        (*vit).hasnormal = true;
    }

    /* Great, now it's time to make all the polygons. */
    for( pit = polylist.begin(); pit != polylist.end(); pit++ ) {
        vector<int>::const_iterator iit;
        VertexList plist;

        for( iit = (*pit).verticies.begin(); iit != (*pit).verticies.end(); iit++ ) {
            plist.push_back( vlist[ (*iit) ] );
        }

        if( (*pit).verticies.size() == 3 ) {
            n->addObject( new Triangle( plist ) );
        } else {
            n->addObject( new ConvexPoly( plist ) );
        }
    }
}

Node *SceneParser::parseNode( void )
{
    VertexList vlist;
    Material *m = 0;
    bool moverride = false;
    Node *n = new Node;
    string curtoken;
    int samplerate = 0;

    n->setName( reader->getNextToken().c_str() );

    while( !reader->isDone() ) {
        curtoken = reader->getNextToken();
        if( curtoken == "endnode" ) break;
        if( curtoken == "translate" ) {
            double x, y, z;
            readTriple( reader, x, y, z );
            n->applyMatrix( buildTranslation( x, y, z ) );
        } else if( curtoken == "scale" ) {
            double x, y, z;
            readTriple( reader, x, y, z );
            n->applyMatrix( buildScale( x, y, z ) );
        } else if( curtoken == "rotatex" ) {
            double a;
            a = atof( reader->getNextToken().c_str() );
            n->applyMatrix( buildRotationX( a ) );
        } else if( curtoken == "rotatey" ) {
            double a;
            a = atof( reader->getNextToken().c_str() );
            n->applyMatrix( buildRotationY( a ) );
        } else if( curtoken == "rotatez" ) {
            double a;
            a = atof( reader->getNextToken().c_str() );
            n->applyMatrix( buildRotationZ( a ) );
        } else if( curtoken == "surface" ) {
            curtoken = reader->getNextToken();
            m = getMaterial( curtoken.c_str() );
            if( !m ) {
                fprintf( stderr, "sceneparser: Material '%s' not found.\n",
                         curtoken.c_str() );
            }
        } else if( curtoken == "surfaceoverride" ) {
            moverride = true;
        } else if( curtoken == "node" ) {
            n->addObject( parseNode() );
        } else if( curtoken == "pointlight" ) {
            n->addObject( new PointLight );
        } else if( curtoken == "instance" ) {
            curtoken = reader->getNextToken();
            Node *in = getNode( curtoken.c_str() );
            if( in ) {
                n->addObject( in );
            } else {
                fprintf( stderr, "sceneparser: Instance '%s' not found.\n",
                                 curtoken.c_str() );
            }
        } else if( curtoken == "verticies" ) {
            (void) reader->getNextToken(); // read {
            while( !reader->isDone() ) {
                double x, y, z;
                curtoken = reader->getNextToken();
                if( curtoken == "}" ) break;
                readTriple( reader, x, y, z ); /* XXX XXX XXX XXX XXX XXX */
                vlist.push_back( Vertex( Point3( x, y, z ) ) );
                (void) reader->getNextToken(); // read }
            }
        } else if( curtoken == "polygons" ) {
            readPolygons( n, vlist );
        } else if( curtoken == "samplerate" ) {
            samplerate = atoi( reader->getNextToken().c_str() );
        }
    }
    if( m ) n->setMaterial( m, moverride );
    if( samplerate ) n->setSampleRate( samplerate );

    return n;
}

